#!/usr/bin/env python

# TPOG Platform Game
# Copyright (C) 2008 Paul Francis Harrison
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

    See http://www.logarithmic.net/pfh/tpog for game-play instructions.


Changes

0.2 - Remove dependance on gdbm (not available under Windows)

0.1 - Initial release

"""

import time, pygame, sys, os.path, marshal


# =============================
# =============================
# Resource loading
# =============================
# =============================

def read_solution_cache():
    global Solution_cache
    try:
        f = open(Solution_cache_filename,'rb')
    except IOError:
        Solution_cache = { }
        return
    Solution_cache = marshal.load(f)
    f.close()

def write_solution_cache():
    global Solution_cache
    f = open(Solution_cache_filename,'wb')
    marshal.dump(Solution_cache,f,1)
    f.close()
    

pygame.init()

Root_dir = os.path.split(__file__)[0]

Solution_cache_filename = os.path.join(Root_dir,'solution_cache.db')
read_solution_cache()

Screen_width = 1000
Screen_height = 700
pygame.display.set_caption('The Problem of Good')
Screen = pygame.display.set_mode((Screen_width, Screen_height))

Big_font = pygame.font.Font(pygame.font.get_default_font(), 60)
Medium_font = pygame.font.Font(pygame.font.get_default_font(), 40)
Little_font = pygame.font.Font(pygame.font.get_default_font(), 20)

Images = {
    'player0' : 'Character Boy.png',
    'player1' : 'Character Horn Girl.png',    
    'block' : 'Plain Block.png',
    'shadow left' : 'Shadow East.png',
    'shadow right' : 'Shadow West.png', 
    'star' : 'Star.png', 
    
    'lift' : 'lift.png',
    'lift up' : 'lift_up.png',
    'button' : 'button.png',
    'button down' : 'button_down.png',
    
    'heart' : 'heart.png',
}

for name in Images.keys():
    Images[name] = pygame.image.load(os.path.join(Root_dir,'image',Images[name]))

for letter in 'ABC':
    surf = Medium_font.render(letter,True,(64,64,64))
    rect = surf.get_rect()
    image = pygame.surface.Surface((101,171)).convert_alpha() # Hmm, ignores srcalpha flag :-(
    image.fill((0,0,0,0))
    image.blit(surf, rect.move(101//2-rect.width//2,4+171-rect.height))
    Images[letter] = image

def write(y,x,align,color,font,text):
    for line in text.split('\n'):
       surf = font.render(line,True,color)
       rect = surf.get_rect()
       Screen.blit(surf, rect.move(x - int(rect.width*align),y))
       y += font.get_height()



# =============================
# =============================
# Game
# =============================
# =============================

class Var(object):
    def __init__(self, state_space, limit):
        self.mult = state_space.limit
        self.limit = limit
        state_space.limit *= limit

    def __call__(self, state):
        return state // self.mult % self.limit

def set(state, *args):
    for i in xrange(0,len(args),2):
        state = state + (args[i+1] - args[i](state)) * args[i+0].mult
    return state

def add(state, *args):
    for i in xrange(0,len(args),2):
        state = state + args[i+1] * args[i+0].mult
    return state

class State_space(object):
    def __init__(self):
        self.limit = 1    

class Game(State_space):
    Motivations0 = [ [10,0 ], [10,  5] ]
    Motivations1 = [ [0 ,10], [ 5, 10] ]

    def __init__(self, blurb, win_comment, board):
        State_space.__init__(self)
        
        self.blurb = blurb
        self.win_comment = win_comment
        self.board = board
        
        self.width = len(board[0])
        self.height = len(board)
        
        self.turn = Var(self,2)
        self.x0 = Var(self, self.width)
        self.y0 = Var(self, self.height)
        self.motive0 = Var(self, len(self.Motivations0))
        
        self.x1 = Var(self, self.width)
        self.y1 = Var(self, self.height)
        self.motive1 = Var(self, len(self.Motivations1))
                
        self.stars = { } # (y,x) -> Var
        
        self.buttons = { } # (y,x) -> id
        self.lifts = { } # (y,x) -> id
        
        for y in xrange(self.height):
            for x in xrange(self.width):
                if board[y][x] == '*':
                    self.stars[(y,x)] = Var(self, 2)
                
                if board[y][x].islower():
                    self.buttons[(y,x)] = board[y][x].upper()
                if board[y][x].isupper():
                    self.lifts[(y,x)] = board[y][x]

        self.score0 = Var(self, len(self.stars)+1) 
        self.score1 = Var(self, len(self.stars)+1)
        
        self.initial_state = 0
        
        for star_var in self.stars.values():
            self.initial_state = set(self.initial_state, star_var, 1)

        for y in xrange(self.height):
            for x in xrange(self.width):
                if board[y][x] == '0':
                    self.initial_state = set(self.initial_state,
                        self.y0, y,
                        self.x0, x
                    )
                if board[y][x] == '1':
                    self.initial_state = set(self.initial_state,
                        self.y1, y,
                        self.x1, x
                    )

    def uid(self):
        return '\n'.join(self.board)

    def relative_score(self, observer, state):
        if self.turn(observer):
            motive = self.Motivations1[self.motive1(observer)]
        else:
            motive = self.Motivations0[self.motive0(observer)]
        return motive[0] * self.score0(state) + motive[1] * self.score1(state)
    
    def block_at(self, y,x):
        return y >= 0 and x >= 0 and y < self.height and x < self.width and \
               self.board[y][x] == '#'    

    def solid_at(self, state, y,x):
        if y < 0 or x < 0 or y >= self.height or x >= self.width or \
           self.board[y][x] == '#' or \
           (y == self.y0(state) and x == self.x0(state)) or \
           (y == self.y1(state) and x == self.x1(state)):
            return True
        
        if (y,x) not in self.lifts:
            return False
        
        pos0 = self.y0(state), self.x0(state)
        if self.buttons.get(pos0,None) == self.lifts[(y,x)]:
            return True
        pos1 = self.y1(state), self.x1(state)
        if self.buttons.get(pos1,None) == self.lifts[(y,x)]:
            return True
        return False
        

    def moves(self, state):
        # In order of decreasing preference (all other things being equal)
        return ('pass', 'left', 'right', 'up', 'love')

    def move(self, state, move):
        if self.turn(state) == 0:
            x = self.x0
            y = self.y0
            motive = self.motive0
            other_x = self.x1
            other_y = self.y1
        else:
            x = self.x1
            y = self.y1
            motive = self.motive1
            other_x = self.x0
            other_y = self.y0
        
        if move == 'love':
            if motive(state)+1 < motive.limit:
                state = add(state,motive,1)
        
        if y(state) < self.height-1 and \
           not self.solid_at(state,y(state)+1,x(state)):
            state = add(state,y,1)        
        else:
            if move == 'left' and \
               not self.solid_at(state,y(state),x(state)-1):
                state = add(state,x,-1)
            if move == 'right' and \
               not self.solid_at(state,y(state),x(state)+1):
                state = add(state,x,1)
            if move == 'up' and \
               not self.solid_at(state,y(state)-1,x(state)) and \
               self.solid_at(state,y(state)+1,x(state)):
                state = add(state,y,-1)

        position = (y(state),x(state))
        if position in self.stars and \
           self.stars[position](state):
            state = set(state,self.stars[position],0)
            if self.turn(state):
                state = add(state,self.score1,1)
            else:
                state = add(state,self.score0,1)
        
        #Lift activated?
        if self.buttons.get( (y(state),x(state)), -1) == \
           self.lifts.get( (other_y(state),other_x(state)), -2):
            state = add(state, other_y, -1)
        
        state = set(state,self.turn,1-self.turn(state))
        
        return state

    def show(self, state):
        render = [ [ ('#' if char == '#' else ' ') 
                     for char in line ] 
                   for line in self.board ]
        
        render[ self.y1(state) ][ self.x1(state) ] = '1'
        render[ self.y0(state) ][ self.x0(state) ] = '0'
        
        for position in self.stars:
            if self.stars[position](state):
                render[ position[0] ][ position[1] ] = '*'
        
        print '='*self.width, state
        for line in render:
            print ''.join(line)
        print '='*self.width

    def draw(self, state, offset_y, offset_x):
        to_draw = [ ]

        positions = ((self.y0(state), self.x0(state)),(self.y1(state), self.x1(state)))
        lifts_active = (self.buttons.get(positions[0],None), self.buttons.get(positions[1],None))
        
        for y in xrange(self.height):
            for x in xrange(self.width):
                if self.block_at(y,x):
                    to_draw.append( (y,x,0,'block') )
                    if self.block_at(y+1,x-1):
                        to_draw.append( (y+1,x-1,1,'shadow left') )
                    if self.block_at(y+1,x+1):
                        to_draw.append( (y+1,x+1,1,'shadow right') )

        for position in self.stars:
            if self.stars[position](state):
                to_draw.append( (position[0],position[1],2,'star') )
        
        for position in self.buttons:
            to_draw.append( (position[0]+1,position[1],5,self.buttons[position].upper()) )
            if position in positions:
                to_draw.append( (position[0],position[1],2,'button down') )
            else:
                to_draw.append( (position[0],position[1],2,'button') )

        for position in self.lifts:
            to_draw.append( (position[0]+1,position[1],5,self.lifts[position]) )
            if self.lifts[position] in lifts_active:
                to_draw.append( (position[0],position[1],2,'lift up') )
            else:
                to_draw.append( (position[0],position[1],2,'lift') )
        
        for y,x,motive,icon in ( (self.y1, self.x1, self.motive1, 'player1'),
                                (self.y0, self.x0, self.motive0, 'player0') ):
            to_draw.append( (y(state),x(state),3,icon) )
            if motive(state) and not self.solid_at(state,y(state)-1,x(state)):
                to_draw.append( (y(state)-2,x(state),4,'heart') )
        
        
        to_draw.sort(key=lambda item: (-item[0],-item[1],item[2]))
        for y, x, z, name in to_draw:
            image = Images[name]
            Screen.blit(image, image.get_rect().move( offset_x+x*100, offset_y+y*41))



def solve(game):
    successors = { } # state -> [ (move, new_state) ]
    
    print 'visit'
    
    # Explore entire reachable state-space    
    order = [ game.initial_state ]
    seen = { game.initial_state : True }
    i = 0
    while i < len(order):
        state = order[i]
        i += 1
        
        successors[state] = [ ]
        successors_seen = { }
        for move in game.moves(state):
            new_state = game.move(state, move)
            if new_state in successors_seen: continue
            successors_seen[new_state] = True
            
            successors[state].append((move, new_state))
            
            #if new_state not in predecessors:
            #    predecessors[new_state] = { }
            #if state not in predecessors[new_state]:
            #    predecessors[new_state][state] = move                
            
            if new_state not in seen:
                order.append(new_state)
                seen[new_state] = True
    
    print len(seen), 'states'
    print 'solve'
    
    #original = state_scores.copy()
    chosen_successor = { }
    state_move = { } # state -> move
    for state in order:
        state_move[state] = successors[state][0][0]
        chosen_successor[state] = successors[state][0][1]
    
    def path_max(observer, state):
        visited = { }
        score = game.relative_score(observer, state)
        best_score = (score, 0)
        while True:
            visited[state] = True
            if state not in chosen_successor: break
            state = chosen_successor[state]
            if state in visited: break
            score = game.relative_score(observer, state)
            best_score = max(best_score, (score, -len(visited)))
        return best_score 
    
    any = True
    # If this halts, it is correct
    # It has always halted... so far
    
    while any:
        print 'iterating'
        any = False
        for state in order:
            player = game.turn(state)
            best_score = None
            for successor in successors[state]:
                #this_score = state_scores[successor[1]][player]
                this_score = path_max(state, successor[1])
                if best_score is None or \
                   this_score > best_score: # or \
                   #(this_score == best_score and successor[0] == state_move[state]):
                    best_score = this_score
                    best = successor
            
            #any = any or state_scores[state] != state_scores[best[1]]
            any = any or best[0] != state_move[state]
            
            state_move[state] = best[0]
            #state_scores[state] = state_scores[best[1]]
            chosen_successor[state] = best[1]
    
    return state_move            


class Solution(object):
    def __init__(self, game):    
        self.game = game
        
        uid = game.uid()
        if uid not in Solution_cache:
            Solution_cache[uid] = solve(game)
            write_solution_cache()
        
        self.state_move = Solution_cache[uid]
        
        self.expectation_cache = { }

    def is_terminal(self, state):
        state2 = self.game.move(state, self.state_move[state])
        state3 = self.game.move(state2, self.state_move[state2])
        return state == state3

    def expectation(self, observer, state):
        key = (observer, state)
        if key not in self.expectation_cache:
            while not self.is_terminal(state):
                state = self.game.move(state, self.state_move[state])
            self.expectation_cache[key] = self.game.relative_score(observer, state)
        return self.expectation_cache[key]

    def score_human(self, observer, history):
        wrongness = 0
        for i in xrange(len(history)):
            if self.game.turn(history[i]) == self.game.turn(observer):
                then_possible = self.expectation(history[i], history[i])
                now_possible = self.expectation(history[i], history[-1])
                loss = then_possible - now_possible
                wrongness = max(wrongness, loss)
        
        current_score = self.game.relative_score(observer, history[-1])
        current_loss = self.expectation(observer, history[0]) - \
                       self.expectation(observer, history[-1])

        deduction = 0
        if wrongness > current_loss:
            deduction = wrongness - current_loss
            
        return current_score, deduction, current_score - deduction

    def play_out(self):
        state = self.game.initial_state
        while True:
            #self.game.show(state)

            Screen.fill((0,196,255))            
            self.game.draw(state, 10, 10)
            pygame.display.flip()
            
            if self.is_terminal(state): break

            time.sleep(0.125)
            
            print self.game.turn(state), self.state_move[state]
            state = self.game.move(state, self.state_move[state])

    def play(self, title, has_prev, has_next):            
        state = self.game.initial_state
        last_state1 = None
        last_state2 = None
        running = True
        keys_down = { }
        
        observer = set(self.game.initial_state,self.game.turn,0)
        
        history = [ ]
        while running:
            history.append(state)
        
            Screen.fill((0,196,255))
            
            score, deduction, true_score = self.score_human(observer, history)
            
            write(20,Screen_width-20,1.0,(255,255,255),Medium_font,
                  title)
            write(20,20,0.0,(255,255,255),Medium_font,
                  'Score: %d' % true_score)
            
            write(Screen_height-45,20,0.0,(255,255,255),Little_font,
                  'Use arrow keys to move and jump, press [L] to fall in love with player 2\n'
                  '[R] replay   ' + 
                  ('[P] previous scenario   ' if has_prev else '') +
                  ('[N] next scenario' if has_next else ''))
            
            #surf = Font.render('Score: %d' % true_score,True,(255,255,255))
            #rect = surf.get_rect()
            #Screen.blit(surf, rect.move(Screen_width//2 - rect.width//2,20))
            
            if deduction:
                write(100,Screen_width//2,0.5,(255,255,255),Little_font,
                      'Lose %d points for being a dick' % deduction +
                      ('\n\nPress [R] to try again' if self.is_terminal(state) else ''))
            
            elif self.is_terminal(state):
                if true_score == self.expectation(observer, self.game.initial_state):    
                    write(100,Screen_width//2,0.5,(255,255,255),Little_font,
                          'Perfect!\n\n' +
                          self.game.win_comment +
                          ('\n\nPress [N] to continue to the next scenario' if has_next else ''))
                else:
                    write(100,Screen_width//2,0.5,(255,255,255),Little_font,
                          'You can do better than that\n\nPress [R] to try again')

            else:
                write(100,Screen_width//2,0.5,(255,255,255),Little_font, self.game.blurb)
                        
            self.game.draw(state, 100, 20)

                #surf = Font.render(
                #    '<<< you lose %d points for being a dick >>>' % deduction,
                #    True,(255,255,255))
                #rect = surf.get_rect()
                #Screen.blit(surf, rect.move(Screen_width//2 - rect.width//2,100))
                        
            pygame.display.flip()
            time.sleep(0.075)
            
            if self.game.turn(state) == 1:
                action = self.state_move[state]
            else:
                if last_state2 == state:
                    events = [ pygame.event.wait() ] + pygame.event.get()
                else:
                    events = pygame.event.get()
                    
                old_keys_down = keys_down.copy()
                keys_seen = { }
                for event in events:
                    if event.type == pygame.QUIT:
                        return 'quit'
                        
                    if event.type == pygame.KEYDOWN:
                        keys_down[event.key] = True
                        keys_seen[event.key] = True
                    if event.type == pygame.KEYUP and event.key in keys_down:
                        del keys_down[event.key]
                        #del keys_seen[event.key]

                keys = keys_seen.copy()
                for key in keys_down:
                    if key in old_keys_down:
                        keys[key] = True
                

                action = 'pass'
                if 276 in keys:
                    action = 'left'
                elif 275 in keys:
                    action = 'right'
                elif 273 in keys:
                    action = 'up'
                elif ord('l') in keys or ord('L') in keys:
                    action = 'love'
                
                if 27 in keys:
                    return 'quit'
                
                if (ord('n') in keys or ord('N') in keys) and has_next:
                    return 'next'
                if (ord('p') in keys or ord('P') in keys) and has_prev:
                    return 'prev'
                if ord('r') in keys or ord('R') in keys:
                    return 'replay'
                
                #if 32 in keys:    
                #    action = self.state_move[state]
                        
            last_state2 = last_state1
            last_state1 = state
            state = self.game.move(state, action)
            
# =============================
# =============================
# Scenarios 
# =============================
# =============================

Games = [
    Game(
    'Collect as many stars as you can!',
    'You have correctly sacrificed the collective\n'
    'good in order to maximize your own reward.', 
    [
     '         ',
     '1        ',
     '##       ',
     '#        ',
     '#   0  * ',
     '#  ######',
     '#       #',
     '# *   * #',
     '#   *   #',
     '#########',
    ]),

    Game(
    'Sometimes cooperation happens by accident.',
    '',
     [
      '         ',
      '         ',
      '         ',
      '      * *',
      '0 1 A#a##',
      '#########',
    ]),
    
    Game(
    'Player 2 has decided to like you.\n\n'
    'This is convenient, because if you can free her\n'
    'she can help you get that star.',
    '',
    [
      '         ',
      '         ',
      '         ',
      '     *   ',
      '     #A1#',
      '     ####',
      '*     ## ',
      '      ## ',
      ' 0 a  ## ',
      '#########',
    ]),

    Game(
    'Your turn.',
    '',
    [
      '         ',
      '   *     ',
      '      0  ',
      '   #A #  ',
      '   ####  ',
      '         ',
      '         ',
      '*   1  a ',
      '# ###### ',
      '#########',
    ]),

     
    Game(
    'Now for a tricky one.',
    'Awww.', 
    [
     '         ',
     '         ',
     '    1    ',
     ' ######  ',
     '    #    ',
     '    #    ',
     '*   # *  ',
     '  0b#*   ',
     'A ####B a',
     '#########',
    ]),

    Game(
    'Final scenario, you lead.',
    'Thanks for playing my game,\nhope you had fun!', 
    [
     '         ',
     '  **     ',
     '         ',
     'a BB 01  ',
     '#######  ',
     '#        ',
     '#       *',
     '#      A#',
     '# b    ##',
     '#########',
    ]),    
]

Solutions = [ Solution(item) for item in Games ]

i = 0
while True:
    result = Solutions[i].play('Scenario %d of %d' % (i+1,len(Solutions)), i > 0, i < len(Solutions)-1)

    if result == 'quit': 
        break
    if result == 'next':
        i += 1
    if result == 'prev':
        i -= 1

